"""
Test the utils functions.
"""

from redbrick.utils import compare_taxonomy
from redbrick.tests.taxonomy2_test import TAXONOMY1, TAXONOMY2
from redbrick.entity.taxonomy2 import Taxonomy2
from redbrick.utils import url_to_image
import cv2  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import numpy as np  # type: ignore


def test_compare() -> None:
    """Test category path with simple taxonomy."""
    Taxonomy1 = Taxonomy2(remote_tax=TAXONOMY1["taxonomy"])

    cat_path = [["object", "car"]]
    assert compare_taxonomy(cat_path, Taxonomy1)

    cat_path2 = [["object", "not correct"]]
    assert not compare_taxonomy(cat_path2, Taxonomy1)

    cat_path3 = [["not correct", "car"]]
    assert not compare_taxonomy(cat_path3, Taxonomy1)

    cat_path4 = [["object"]]
    assert not compare_taxonomy(cat_path4, Taxonomy1)

    cat_path5 = [["object", "car", "not correct"]]
    assert not compare_taxonomy(cat_path5, Taxonomy1)


def test_url_to_image() -> None:
    """Test the conversion of a url to image data."""
    url = "https://asia.olympus-imaging.com/content/000107506.jpg"
    local = "redbrick/tests/000107506.jpg"

    # get local image
    img_local = cv2.imread(local)
    img_local = np.flip(img_local, axis=2)

    # get remote image
    img_remote = url_to_image(url)

    comparison = img_local == img_remote

    assert comparison.all()


def test_url_to_image_invalid() -> None:
    """Test the utility function with invalid urls"""
    url = "https://not-a-real-url.jpg"
    url2 = "http://127.0.0.1:8080/not-real.jpg"

    # first image
    try:
        url_to_image(url)
        assert False
    except:
        assert True

    # second image
    try:
        url_to_image(url2)
        assert False
    except:
        assert True


if __name__ == "__main__":
    test_url_to_image_invalid()



"""
Representation of a bounding box labels
"""

from dataclasses import dataclass
from redbrick.entity.taxonomy import TaxonomyEntry
from redbrick.utils import compare_taxonomy
from redbrick.entity.taxonomy2 import Taxonomy2
from typing import List, Union, Dict, Any, Tuple, Optional
import numpy as np  # type: ignore
import os
import json
from tqdm import tqdm  # type: ignore
import copy
import uuid
import matplotlib.patches as patches  # type: ignore
import copy


@dataclass
class ImageBoundingBoxEntry:

    xnorm: float
    ynorm: float
    wnorm: float
    hnorm: float
    classname: List[List[str]]


@dataclass
class VideoBoundingBoxEntry:

    xnorm: float
    ynorm: float
    wnorm: float
    hnorm: float
    classname: List[List[str]]
    labelid: str
    trackid: str
    frameindex: int
    keyframe: bool
    end: bool


class VideoBBoxLabel:
    def __init__(
        self,
        labelid: str,
        trackid: str,
        xnorm: float,
        ynorm: float,
        hnorm: float,
        wnorm: float,
        frameindex: int,
        keyframe: bool,
        end: bool,
        category: List[List[str]],
    ) -> None:
        self.labelid = labelid
        self.trackid = trackid
        self.xnorm = xnorm
        self.ynorm = ynorm
        self.hnorm = hnorm
        self.wnorm = wnorm
        self.frameindex = frameindex
        self.keyframe = keyframe
        self.end = end
        self.category = category

    def copy(self) -> "VideoBBoxLabel":
        return VideoBBoxLabel(
            self.labelid,
            self.trackid,
            self.xnorm,
            self.ynorm,
            self.hnorm,
            self.wnorm,
            self.frameindex,
            self.keyframe,
            self.end,
            self.category,
        )

    @classmethod
    def from_dict(cls, val: Any) -> "VideoBBoxLabel":
        return cls(
            labelid=val.labelid,
            trackid=val.trackid,
            xnorm=val.xnorm,
            ynorm=val.ynorm,
            wnorm=val.wnorm,
            hnorm=val.hnorm,
            frameindex=val.frameindex,
            end=val.end,
            category=val.classname[0][-1],
            keyframe=val.keyframe,
        )


@dataclass
class BaseBoundingBox:
    labels: Union[List[ImageBoundingBoxEntry], List[VideoBoundingBoxEntry]]

    def compare_tax(self, taxonomy: Taxonomy2) -> Tuple[bool, List[List[str]]]:
        """Compare the labels category to taxonomy."""
        raise NotImplementedError()

    def __getitem__(
        self, index: int
    ) -> Union[ImageBoundingBoxEntry, VideoBoundingBoxEntry]:
        """Get item."""
        return self.labels[index]

    def __iter__(self) -> "BaseBoundingBox":
        """Label iterator."""
        self.it = 0
        return self

    def __next__(self) -> Union[ImageBoundingBoxEntry, VideoBoundingBoxEntry]:
        """Next iterator."""
        if self.it < len(self.labels):
            val = self.labels[self.it]
            self.it += 1
            return val
        else:
            raise StopIteration


class ImageBoundingBox(BaseBoundingBox):
    """Bounding box for a single image."""

    def __init__(self, labels: List[Any]) -> None:
        """Constructor."""
        self.labels: List[ImageBoundingBoxEntry]

        # Parse the remote labels, and save as entries
        entries = []
        for label in labels:

            if "bbox2d" not in label:
                raise ValueError(
                    "Incorrect format for labels entry ImageBoundingBox. bbox2d field is missing!"
                )

            label_ = label["bbox2d"]
            entry = ImageBoundingBoxEntry(
                xnorm=label_["xnorm"],
                ynorm=label_["ynorm"],
                wnorm=label_["wnorm"],
                hnorm=label_["hnorm"],
                classname=label["category"],
            )
            entries.append(entry)

        self.labels = entries

    def compare_taxonomy(self, taxonomy: Taxonomy2) -> Tuple[bool, List[List[str]]]:
        """Compare the label classes with the taxonomy."""
        # Iterate over each label
        for label in self.labels:
            check_ = compare_taxonomy(label.classname, taxonomy)
            if not check_:
                return False, label.classname

        return True, [["null", "null"]]

    def show(self, ax: Any, width: int, height: int) -> None:
        """Show the image bbox."""
        for label in self.labels:
            x = label.xnorm * width
            y = label.ynorm * height
            h = label.hnorm * height
            w = label.wnorm * width
            rect = patches.Rectangle(
                xy=(x, y),
                width=w,
                height=h,
                edgecolor=(5 / 256, 4 / 256, 170 / 256),
                facecolor=(5 / 256, 4 / 256, 170 / 256, 0.3),
            )
            ax.add_patch(rect)
            ax.text(
                x,
                y - 1,
                label.classname[0][-1],
                fontsize=10,
                color=(5 / 256, 4 / 256, 170 / 256),
            )

    def __str__(self) -> str:
        """String representation."""
        labels = []
        for label in self.labels:
            entry: Dict[Any, Any] = {}
            entry["category"] = label.classname
            entry["attributes"] = []
            entry["labelid"] = str(uuid.uuid4())
            entry["bbox2d"] = {
                "xnorm": label.xnorm,
                "ynorm": label.ynorm,
                "hnorm": label.hnorm,
                "wnorm": label.wnorm,
            }

            labels.append(entry)

        output = {"items": [{"url": "throwaway", "labels": labels}]}

        return json.dumps(output)


class VideoBoundingBox(BaseBoundingBox):
    """Bounding box for a single video."""

    def __init__(self, labels: List[Any]) -> None:
        """Constructor."""
        self.labels: List[VideoBoundingBoxEntry]
        self.interpolated_labels: Optional[List[List[VideoBBoxLabel]]] = None

        # parse the remote labels, and store entries
        entries = []
        for label in labels:

            # Fill out values if not provided
            if not "trackid" in label:
                label["trackid"] = str(uuid.uuid4())
            if not "end" in label:
                label["end"] = True
            if not "keyframe" in label:
                label["keyframe"] = True
            if not "labelid" in label:
                label["labelid"] = str(uuid.uuid4())

            entry = VideoBoundingBoxEntry(
                xnorm=label["bbox2d"]["xnorm"],
                ynorm=label["bbox2d"]["ynorm"],
                wnorm=label["bbox2d"]["wnorm"],
                hnorm=label["bbox2d"]["hnorm"],
                classname=label["category"],
                labelid=label["labelid"],
                trackid=label["trackid"],
                frameindex=label["frameindex"],
                keyframe=label["keyframe"],
                end=label["end"],
            )
            entries.append(entry)

        self.labels = entries

    def show(
        self, ax: Any, width: int, height: int, frameindex: int, num_frames: int
    ) -> List[Any]:
        """Show the video bbox."""
        # interpolate labels if not already done
        if not self.interpolated_labels:
            frame_labels = self.interpolate_labels(num_frames=num_frames)
            self.interpolated_labels = frame_labels

        ims = []
        for label in self.interpolated_labels[frameindex]:
            x = label.xnorm * width
            y = label.ynorm * height
            h = label.hnorm * height
            w = label.wnorm * width
            rect = patches.Rectangle(
                xy=(x, y),
                width=w,
                height=h,
                edgecolor=(5 / 256, 4 / 256, 170 / 256),
                facecolor=(5 / 256, 4 / 256, 170 / 256, 0.3),
                animated=True,
            )
            im1 = ax.add_patch(rect)
            im2 = ax.text(
                x,
                y - 1,
                label.category[0][-1],
                fontsize=10,
                color=(5 / 256, 4 / 256, 170 / 256),
                animated=True,
            )
            ims += [im1, im2]

        return ims

    def compare_taxonomy(self, taxonomy: Taxonomy2) -> Tuple[bool, List[List[str]]]:
        """Ensure labels abide by taxonomy."""
        # Iterate over each frame
        for label in self.labels:
            check = compare_taxonomy(category_path=label.classname, taxonomy=taxonomy)

            if not check:
                return check, label.classname

        return True, [["null", "null"]]

    def __str__(self) -> str:
        """String representation of self."""
        labels = []

        for label in self.labels:
            entry: Dict[Any, Any] = {}
            entry["category"] = label.classname
            entry["attributes"] = []
            entry["labelid"] = label.labelid
            entry["bbox2d"] = {
                "xnorm": label.xnorm,
                "ynorm": label.ynorm,
                "hnorm": label.hnorm,
                "wnorm": label.wnorm,
            }
            entry["frameindex"] = label.frameindex
            entry["keyframe"] = label.keyframe
            entry["trackid"] = label.trackid
            entry["end"] = label.end
            labels.append(entry)

        output = {"labels": labels}
        return json.dumps(output)

    def interpolate_labels(self, num_frames: int) -> List[List[VideoBBoxLabel]]:
        """Interpolate the frames and return interpolated object."""
        # Write labels for each frame here
        # Write labels for each frame here
        temp: Dict[Any, Any] = {}
        for label in self.labels:
            if label.trackid in temp:
                temp[label.trackid].append(label)
            else:
                temp[label.trackid] = [label]

        def interpolate(
            start: VideoBBoxLabel, frameindex: int, end: Optional[VideoBBoxLabel],
        ) -> "VideoBBoxLabel":
            if not end:
                result = start.copy()
                result.keyframe = False
                return result
            assert frameindex > start.frameindex
            assert frameindex < end.frameindex
            before = frameindex - start.frameindex
            after = end.frameindex - frameindex
            total = end.frameindex - start.frameindex
            temp_ = start.copy()
            assert before > 0
            assert after > 0
            assert total > 0
            temp_.xnorm = after / total * start.xnorm + before / total * end.xnorm
            temp_.ynorm = after / total * start.ynorm + before / total * end.ynorm
            temp_.hnorm = after / total * start.hnorm + before / total * end.hnorm
            temp_.wnorm = after / total * start.wnorm + before / total * end.wnorm
            temp_.keyframe = False
            return temp_

        def interpolate2(
            keyframes: List[Optional[VideoBBoxLabel]], frameindex: int
        ) -> "Optional[VideoBBoxLabel]":
            # keyframes is a list where length = numFrames, keyframes[label.frameindex]=label
            # [ None, label, None, None, label]
            temp = keyframes[frameindex]
            if temp and temp.keyframe:
                return keyframes[frameindex]
            start: Optional[VideoBBoxLabel] = None

            for ii in reversed(range(0, frameindex)):
                label = keyframes[ii]
                if label and label.keyframe:
                    if label.end:
                        return None
                    start = label
                    break

            if not start:
                return None

            end: Optional[VideoBBoxLabel] = None
            for ii in range(frameindex, len(keyframes), 1):
                label = keyframes[ii]
                if label and label.keyframe:
                    end = label
                    break

            return interpolate(start, frameindex, end)

        def expand_list_of_keyframes(
            keyframes: List[VideoBBoxLabel], num_frames: int
        ) -> List[Optional[VideoBBoxLabel]]:
            # create a sparse list of keyframe labels of same length as num_frames
            mylist: List[Optional[VideoBBoxLabel]] = [None] * num_frames
            for label in keyframes:
                mylist[label.frameindex] = label

            return mylist

        temp2 = {}  # trackid -> full list of labels and keyframes (and None)
        for trackid, labellist in temp.items():
            temp45 = [VideoBBoxLabel.from_dict(label) for label in labellist]
            temp46 = expand_list_of_keyframes(temp45, num_frames)
            output: List[Optional[VideoBBoxLabel]] = [None] * num_frames

            for index, _ in enumerate(output):
                output[index] = interpolate2(temp46, index)

            temp2[trackid] = output

        framelabels: List[List[VideoBBoxLabel]] = []
        for index in range(num_frames):
            current_frame_labels = []
            for labellist in temp2.values():
                templable = labellist[index]
                if templable:
                    current_frame_labels.append(templable)
            framelabels.append(current_frame_labels)

        return framelabels



"""
Base class for datapoints.
"""
from dataclasses import dataclass, field
from typing import List, Optional, Union, Dict, Any
import datetime


@dataclass
class BaseDatapoint:
    """Base class representation of a datapoint."""

    org_id: str
    label_set_name: str
    taxonomy: Dict[str, int]
    task_type: str
    remote_labels: List[Any]



"""Getting data from redbrick api."""

import redbrick
import json
from dataclasses import dataclass, asdict
from typing import Dict, List, Any, Optional, Union, Tuple
import requests
import numpy as np  # type: ignore
import cv2  # type: ignore
from redbrick.utils import url_to_image
from .api_base import RedBrickApiBase
from .entity.custom_group import CustomGroup
from .entity.datapoint import Image
from .entity.datapoint import Video
from .entity.task import Task
from .entity.taxonomy2 import Taxonomy2


class RedBrickApi(RedBrickApiBase):
    """Implement Abstract API."""

    def __init__(self, cache: bool = False, custom_url: Optional[str] = None) -> None:
        """Construct RedBrickApi."""
        self.client = redbrick.client.RedBrickClient()
        if custom_url:
            self.url = custom_url
        elif self.client.custom_url:
            self.url = self.client.custom_url
        else:
            self.url = "https://redbrick-prod-1.herokuapp.com/graphql/"

    def get_datapoint_ids(
        self, org_id: str, label_set_name: str
    ) -> Tuple[List[str], CustomGroup]:
        """Get a list of datapoint ids in labelset."""
        query_string = """
            query ($orgId:UUID!, $name:String!) {
                customGroup(orgId: $orgId, name: $name) {
                    datapoints(skip:0, first:-1) {
                        dpId
                    }
                    taskType,
                    dataType
                    taxonomy {
                    name
                    version
                    categories {
                        name
                        children {
                        name
                        classId
                        children {
                            name
                            classId
                            children {
                            name
                            classId
                            }
                        }
                        }
                    }
                    }
                }
            }
        """
        query_variables = {"orgId": org_id, "name": label_set_name}
        query = dict(query=query_string, variables=query_variables)

        result = self._execute_query(query)

        all_dp_ids = [dp["dpId"] for dp in result["customGroup"]["datapoints"]]
        custom_group = CustomGroup(
            result["customGroup"]["taskType"],
            result["customGroup"]["dataType"],
            result["customGroup"]["taxonomy"],
        )
        return all_dp_ids, custom_group

    def get_datapoint(
        self,
        org_id: str,
        label_set_name: str,
        dp_id: str,
        task_type: str,
        taxonomy: dict,
    ) -> Union[Image, Video]:
        """Get all relevant information related to a datapoint."""
        query_string = """
            query ($orgId: UUID!, $dpId: UUID!, $name:String!) {
                labelData(orgId: $orgId, dpId: $dpId, customGroupName: $name) {
                    blob
                    dataPoint {
                        items:items(presigned:true)
                        items_not_signed:items(presigned:false)
                        name
                    },
                    taskType,
                    dataType, 
                    labels {
                      category,
                      labelid, 
                      frameclassify, 
                      frameindex,
                      trackid, 
                      keyframe, 
                      end,
                      bbox2d {
                          xnorm, 
                          ynorm, 
                          wnorm, 
                          hnorm,
                      }
                    }
                }
            }
        """
        # EXECUTE THE QUERY
        query_variables = {"orgId": org_id, "name": label_set_name, "dpId": dp_id}
        query = dict(query=query_string, variables=query_variables)
        result = self._execute_query(query)

        # IMAGE DATA
        if result["labelData"]["dataType"] == "IMAGE":
            # Parse result
            signed_image_url = result["labelData"]["dataPoint"]["items"][0]
            unsigned_image_url = result["labelData"]["dataPoint"]["items_not_signed"][0]
            labels = json.loads(result["labelData"]["blob"])["items"][0]["labels"]
            image_data = url_to_image(signed_image_url)  # Get image array

            dpoint = Image(
                org_id=org_id,
                label_set_name=label_set_name,
                taxonomy=taxonomy,
                remote_labels=labels,
                dp_id=dp_id,
                image_url=signed_image_url,
                image_url_not_signed=unsigned_image_url,
                image_data=image_data,
                task_type=task_type,
            )
            return dpoint

        # VIDEO DATA
        else:
            # Parse the result
            items = result["labelData"]["dataPoint"]["items"]
            items_not_signed = result["labelData"]["dataPoint"]["items_not_signed"]
            labels = result["labelData"]["labels"]
            with open("result.json", "w+") as file:
                json.dump(result, file, indent=2)
            name = result["labelData"]["dataPoint"]["name"]
            dpoint_vid = Video(
                org_id=org_id,
                label_set_name=label_set_name,
                taxonomy=taxonomy,
                task_type=task_type,
                remote_labels=labels,
                dp_id=dp_id,
                video_name=name,
                items_list=items,
                items_list_not_signed=items_not_signed,
            )
            return dpoint_vid

    def tasksToLabelRemote(self, orgId, projectId, stageName, numTasks) -> List[Task]:
        """Get remote labeling tasks."""
        query_string = """
        query ($orgId:UUID!, $projectId:UUID!, $stageName:String!, $numTasks:Int!){
            tasksToLabelRemote(orgId:$orgId, projectId:$projectId, stageName:$stageName, numTasks:$numTasks) {
                stageName, 
                taskId,
                subName,
                modelName,
                createdAt, 
                dpId, 
                taxonomy {
                    name, 
                    version, 
                    categories {
                        name, 
                        children {
                            name, 
                            classId,
                            children {
                                name, 
                                classId,
                                children {
                                    name,
                                    classId
                                }
                            }
                        }
                    }
                }
                datapoint {
                    items(presigned:false), 
                    itemsPresigned,                    
                } 
            }
        }
        """
        query_variables = {
            "orgId": orgId,
            "projectId": projectId,
            "stageName": stageName,
            "numTasks": numTasks,
        }
        query = dict(query=query_string, variables=query_variables)
        result = self._execute_query(query)
        result = result["tasksToLabelRemote"]

        # Get stage information
        stage_info = self.get_stage(orgId, projectId, stageName)

        tasks = []
        for res in result:
            # Create a task object and return
            task = Task(
                org_id=orgId,
                project_id=projectId,
                stage_name=stageName,
                task_id=res["taskId"],
                dp_id=res["dpId"],
                sub_name=res["subName"],
                taxonomy=res["taxonomy"],
                items_list=res["datapoint"]["items"],
                items_list_presigned=res["datapoint"]["itemsPresigned"],
                task_data_type=stage_info["outputType"],
            )
            tasks.append(task)

        return tasks

    def putTaskData(
        self,
        org_id,
        project_id,
        dp_id,
        stage_name,
        sub_name,
        task_data,
        taxonomy_name,
        taxonomy_version,
        td_type,
        augmentdata=None,
    ) -> None:
        """Put task data for a labeling task."""
        query_string = """
        mutation($orgId:UUID!, $dpId:UUID!, $projectId:UUID!, $stageName:String!, $subName:String!, $taskData:JSONString!, $taxonomyName: String!, $taxonomyVersion: Int!, $tdType: TaskDataType!) {
            putTaskData(orgId:$orgId, dpId:$dpId, projectId: $projectId, stageName:$stageName, subName:$subName, taskData:$taskData, taxonomyName:$taxonomyName, taxonomyVersion: $taxonomyVersion, tdType:$tdType) {
                ok
            }
        }
        """
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "dpId": dp_id,
            "stageName": stage_name,
            "subName": sub_name,
            "taskData": task_data,
            "taxonomyName": taxonomy_name,
            "taxonomyVersion": taxonomy_version,
            "tdType": td_type,
        }
        query = dict(query=query_string, variables=query_variables)
        self._execute_query(query)

    def putRemoteLabelingTask(self, finishedTask) -> None:
        """Put remote labeling task to backend."""
        query_string = """
        mutation($finishedTask: RemoteLabelingTaskInput!) {
            putRemoteLabelingTask(finishedTask: $finishedTask) {
                ok
            }
        }
        """

        query_variables = {"finishedTask": finishedTask}
        query = dict(query=query_string, variables=query_variables)
        self._execute_query(query)

    def get_stage(self, org_id, project_id, stage_name) -> Dict[Any, Any]:
        """Get stage information."""
        query_string = """
            query($orgId: UUID!, $projectId: UUID!, $stageName: String!) {
                stage(orgId: $orgId, projectId: $projectId, stageName: $stageName) {
                    inputType,
                    outputType,
                    outputTaxonomyName, 
                    outputTaxonomyVersion
                }
            }
        """
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
        }
        query = dict(query=query_string, variables=query_variables)
        result = self._execute_query(query)
        return result["stage"]

    def get_num_remote_tasks(self, org_id, project_id, stage_name) -> int:
        """Get stage information."""
        query_string = """
        query($orgId: UUID!, $projectId: UUID!, $stageName: String!) {
            stageStat(orgId: $orgId, projectId: $projectId, stageName: $stageName) {
                currentTaskCount
            }
        }
        """
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
        }
        query = dict(query=query_string, variables=query_variables)
        result = self._execute_query(query)
        return int(result["stageStat"]["currentTaskCount"])

    def get_taxonomy(self, orgId: str, name: str, version: int) -> Taxonomy2:
        """Get a taxonomy info."""
        query_string = """
        query ($orgId: UUID!, $name: String!, $version: Int!) {
            taxonomy(orgId: $orgId, name: $name, version: $version) {
                name, 
                version, 
                categories {
                    name, 
                    children {
                        name, 
                        classId,
                        children {
                            name, 
                            classId,
                            children {
                                name,
                                classId
                            }
                        }
                    }
                }
            }
        }
        """

        query_variables = {
            "orgId": orgId,
            "name": name,
            "version": version,
        }
        query = dict(query=query_string, variables=query_variables)
        result = self._execute_query(query)

        Tax = Taxonomy2(remote_tax=result["taxonomy"])
        return Tax

    def _execute_query(self, query: Dict) -> Any:
        """Execute a graphql query."""
        headers = {"ApiKey": self.client.api_key}
        try:
            response = requests.post(self.url, headers=headers, json=query)
            res = {}
            if "errors" in response.json():
                raise ValueError(response.json()["errors"][0]["message"])
            elif "data" in response.json():
                res = response.json()["data"]
            else:
                res = response.json()
            return res
        except ValueError:
            print(response.content)
            print(response.status_code)
            raise ValueError



from .classify import ImageClassify, VideoClassify
from .bbox import (
    ImageBoundingBox,
    VideoBoundingBox,
    ImageBoundingBoxEntry,
    VideoBoundingBoxEntry,
)
from .segmentation import ImageSegmentation



"""
Base labelset.
"""

from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any


@dataclass
class LabelsetBase:
    org_id: str
    labelset_name: str
    task_type: str
    data_type: str
    taxonomy: Dict[str, int]
    dp_ids: List[str] = field(init=False)

    def __getitem__(self, index: int) -> Any:
        """Get a single dp."""
        raise NotImplementedError()



"""
Base class for export controller.
"""
from redbrick.labelset.labelset_base import LabelsetBase
from dataclasses import dataclass, field
import datetime
import os
from typing import Optional
import uuid


@dataclass
class ExportBase:
    """Base class for export controller."""

    format: str
    labelset: LabelsetBase
    cache_dir: str = field(init=False)
    export_dir: str = field(init=False)

    def __post_init__(self) -> None:
        """Run after init."""
        salt = uuid.uuid4()
        self.cache_dir = "RB_Export_%s_%s" % (self.format, salt)
        self.export_dir = "RB_Export_%s_%s" % (self.format, salt)

    def export(self) -> None:
        """Export the data and labels in the correct format."""
        raise NotImplementedError()

    def cache(self) -> None:
        """Cache the data and labels."""
        os.mkdir(self.cache_dir)

        if self.labelset.task_type == "SEGMENTATION":
            self.cache_segmentation()
        elif self.labelset.task_type == "BBOX":
            self.cache_bbox()
        elif self.labelset.task_type == "CLASSIFY":
            self.cache_classify()

    def cache_bbox(self) -> None:
        """Cache image bounding boxes."""
        raise NotImplementedError()

    def cache_segmentation(self) -> None:
        """Cache image segmentation."""
        raise NotImplementedError

    def cache_classify(self) -> None:
        """Cache image classification."""
        raise NotImplementedError



"""
Test the remote label module.
"""

import redbrick


def test_init() -> None:
    """Test simple init."""
    assert True
    return

    org_id = "52c624b2-b7eb-4196-b5af-204b79aa3f7e"
    project_id = "8c1a1b99-6ca3-4e40-9627-9f1652076e8c"
    stage_name = "LABEL"
    api_key = "w0jS8yw_KDgrfXflBzNEcOrCPm-PU3inOTJ-3niYsI8"
    redbrick.init(api_key=api_key)

    remote_label = redbrick.remote_label.RemoteLabel(
        org_id=org_id, project_id=project_id, stage_name=stage_name
    )

    assert remote_label.taxonomy.name == "DEFAULT::YoloThings"
    assert remote_label.task_type == "IMAGE_BBOX"



from typing import List, Optional
import json
from .taxonomy import TaxonomyEntry
from typing import Dict, Any


class CustomGroup:
    def __init__(
        self, task_type: str, data_type: str, taxonomy: Dict[Any, Any]
    ) -> None:
        """Construct a CustomGroup object."""
        self.task_type = task_type
        self.data_type = data_type

        tax_map: Dict[str, int] = {}
        self.trav_tax(taxonomy["categories"][0], tax_map)
        self.taxonomy = tax_map

    def trav_tax(self, taxonomy: Dict[Any, Any], tax_map: Dict[str, int]) -> None:
        """Traverse the taxonomy tree structure, and fill the taxonomy mapper object."""
        children = taxonomy["children"]
        if len(children) == 0:
            return

        for child in children:
            tax_map[child["name"]] = child["classId"]
            self.trav_tax(child, tax_map)



"""Abstract interface to RedBrick API."""

from typing import List, Union
from abc import ABC, abstractmethod
from redbrick.entity.datapoint import Image, Video


class RedBrickApiBase(ABC):
    """Abstract interface to RedBrick API."""

    @abstractmethod
    def get_datapoint_ids(self, org_id: str, label_set_name: str) -> List[str]:
        """Get a list of datapoint ids in labelset."""

    @abstractmethod
    def get_datapoint(
        self,
        org_id: str,
        label_set_name: str,
        dp_id: str,
        task_type: str,
        taxonomy: dict,
    ) -> Union[Image, Video]:
        """Get all relevant information related to a datapoint."""



"""
Utility functions.
"""

import requests
import numpy as np  # type: ignore
import cv2  # type: ignore
from redbrick.entity.taxonomy2 import Taxonomy2
from typing import List


def url_to_image(url: str) -> np.ndarray:
    """Get image data from url."""
    # Download the image, convert it to a NumPy array, and then read
    # it into OpenCV format

    resp = None

    try:
        resp = requests.get(url, stream=True)
        resp.raw.decode_content = True

        # check for errors
        if not resp.status_code == 200:
            raise Exception("Not able to access data at %s url" % (url))
    except Exception as err:
        raise Exception("%s. Not able to access data at %s url" % (str(err), url))

    image = np.asarray(bytearray(resp.raw.read()), dtype="uint8")
    image = cv2.imdecode(image, cv2.IMREAD_COLOR)
    # cv2 returns a BGR image, need to convert to RGB
    # the copy operation makes the memory contiguous for tensorify-ing
    return np.flip(image, axis=2).copy()


def compare_taxonomy(category_path: List[List[str]], taxonomy: Taxonomy2) -> bool:
    """Check if the category_path is valid for taxonomy."""
    tax_obj = taxonomy.taxonomy["categories"]

    for idx, cat in enumerate(category_path[0]):
        # Iterate through the tax obj
        for elem in tax_obj:
            if cat == elem["name"]:
                tax_obj = elem["children"]

                # Make sure this is the last category in path, and last elem in tax tree
                if len(tax_obj) == 0 and idx == len(category_path[0]) - 1:
                    return True

    return False



# mypy: ignore-errors

"""
Test the classify module.
"""
from redbrick.entity.label import VideoClassify
from .taxonomy2_test import TAXONOMY1
from redbrick.entity.taxonomy2 import Taxonomy2


VID_CLASS_1 = [
    {"category": [["object", "person"]], "frameindex": 0, "frameclassify": True},
    {"category": [["object", "person"]], "frameindex": 1, "frameclassify": True},
    {"category": [["object", "car"]], "frameindex": 2, "frameclassify": True},
]

VID_CLASS_2 = [
    {"category": [["object", "wrong"]], "frameindex": 0, "frameclassify": True},
    {"category": [["object", "person"]], "frameindex": 1, "frameclassify": True},
    {"category": [["object", "car"]], "frameindex": 2, "frameclassify": True},
]


def test_video_classify_init():
    """Test video classify init."""

    video_classify = VideoClassify(remote_labels=VID_CLASS_1)
    assert len(video_classify.labels) == 3


def test_video_classify_compare_taxonomy():
    """Test compare taxonomy."""
    tax = Taxonomy2(remote_tax=TAXONOMY1["taxonomy"])
    video_classify = VideoClassify(remote_labels=VID_CLASS_1)

    check, _ = video_classify.compare_taxonomy(taxonomy=tax)
    assert check


def test_video_classify_compare_taxonomy_wrong():
    """Incorrect taxonomy."""
    tax = Taxonomy2(remote_tax=TAXONOMY1["taxonomy"])
    video_classify = VideoClassify(remote_labels=VID_CLASS_2)

    check, fail = video_classify.compare_taxonomy(taxonomy=tax)
    assert not check
    assert fail == [["object", "wrong"]]



"""
Class representation of a taxonomy.
"""
from typing import Dict, Any, List


class Taxonomy2:
    """Class representation of a taxonomy."""

    def __init__(self, remote_tax: Dict[Any, Any]) -> None:
        """Constructor."""
        self.taxonomy = remote_tax
        self.name = remote_tax["name"]

        # Create class/id map objects
        class_id_map: Dict[str, int] = {}
        id_class_map: Dict[int, str] = {}
        self.class_id_map(
            tax=remote_tax["categories"][0]["children"],
            class_id_map=class_id_map,
            id_class_map=id_class_map,
        )
        self.taxonomy_class_id_map: Dict[str, int] = class_id_map
        self.taxonomy_id_class_map: Dict[int, str] = id_class_map

    def class_id_map(
        self, tax: List[Any], class_id_map: Dict[str, int], id_class_map: Dict[int, str]
    ) -> None:
        """Creates class/id maps recursively, vars are passed by reference."""
        if len(tax) == 0:
            return

        for elem in tax:
            class_id_map[elem["name"]] = elem["classId"]
            id_class_map[elem["classId"]] = elem["name"]
            self.class_id_map(elem["children"], class_id_map, id_class_map)

        return



"""
Class for controlling video exports.
"""
from .export_base import ExportBase
from dataclasses import dataclass
from typing import List, Optional, Dict, Any
import os
import json
import cv2  # type: ignore
import shutil
import numpy as np  # type: ignore
from termcolor import colored
from tqdm import tqdm  # type: ignore
from redbrick.entity.datapoint import Image
from redbrick.utils import url_to_image


@dataclass
class ExportVideo(ExportBase):
    """Handle video exports."""

    def export(self) -> None:
        """Export the video and labels."""
        print(colored("[INFO]:", "blue"), "Cacheing labels, and data...")
        self.cache()

    def cache(self) -> None:
        """Cache the video and labels."""
        ExportBase.cache(self)

    def cache_bbox(self) -> None:
        """Cache video bounding boxes."""
        # Create YOLO style meta data files
        num_classes = len(list(self.labelset.taxonomy.keys()))
        train_file = "train.txt"
        names_file = "names.txt"
        data_file = "obj.data"
        backup_dir = "backup/"

        # obj.data
        with open(self.cache_dir + "/" + data_file, "w+") as file:
            file.write("classes = %s" % num_classes)
            file.write("\n")
            file.write("train = %s" % ("data" + "/" + train_file))
            file.write("\n")
            file.write("names = %s" % ("data" + "/" + names_file))
            file.write("\n")
            file.write("backup = %s" % backup_dir)

        # names.txt
        with open(self.cache_dir + "/" + names_file, "w+") as file:
            class_names = list(self.labelset.taxonomy.keys())
            for name in class_names:
                file.write(name + "\n")

        # obj_train_data/
        os.mkdir(os.path.join(self.cache_dir, "obj_train_data"))
        taxonomy_mapper = {
            name: idx for idx, name in enumerate(list(self.labelset.taxonomy.keys()))
        }
        image_filepaths = []

        for i in range(len(self.labelset.dp_ids)):
            print(
                colored("[INFO]:", "blue"),
                "Exporting %s/%s video(s)" % (i + 1, len(self.labelset.dp_ids)),
            )

            # Create tqdm loader
            dp = self.labelset.__getitem__(i)
            pbar = tqdm(total=len(dp.items_list))

            # Interpolate the video labels, and get per frame labels
            framelabels = dp.labels.interpolate_labels(num_frames=len(dp.items_list))

            # Iterate through video frames
            for idx, item in enumerate(dp.items_list):
                pbar.update(1)
                dp_entry = {}
                dp_entry["url"] = dp.items_list_not_signed[idx]
                dp_entry["labels"] = []

                # write image data to file
                image_filepath = os.path.join(
                    self.cache_dir,
                    "obj_train_data",
                    str(
                        dp.items_list_not_signed[idx]
                        .replace("/", "_")
                        .replace(":", "_")
                    ),
                )
                image_filepaths.append(
                    os.path.join(
                        "obj_train_data",
                        str(
                            dp.items_list_not_signed[idx]
                            .replace("/", "_")
                            .replace(":", "_")
                        ),
                    )
                )

                frame_data = np.flip(url_to_image(item), axis=2)
                cv2.imwrite(image_filepath, frame_data)

                # create the label file name
                filext_idx = image_filepath.rfind(".")
                if not filext_idx:
                    filext_idx = -1
                label_filepath = image_filepath[0:filext_idx] + ".txt"

                # write labels to the txt file
                frame_labels = framelabels[idx]

                # write labels to the txt file
                with open(label_filepath, "w+") as file:
                    for label in frame_labels:
                        class_idx = taxonomy_mapper[label.category]
                        file.write(
                            "%d %.6f %.6f %.6f %.6f \n"
                            % (
                                class_idx,
                                label.xnorm,
                                label.ynorm,
                                label.wnorm,
                                label.hnorm,
                            )
                        )
            pbar.close()

        # create train.txt file
        with open(os.path.join(self.cache_dir, train_file), "w+") as file:
            for filename in image_filepaths:
                file.write(filename + "\n")

    def cache_classify(self) -> None:
        """Cache video classification."""
        output = []
        for i in range(len(self.labelset.dp_ids)):
            print(
                colored("[INFO]:", "blue"),
                "Exporting %s/%s video(s)" % (i + 1, len(self.labelset.dp_ids)),
            )

            dp = self.labelset.__getitem__(i)
            video_name = dp.video_name
            labels = dp.labels

            # Loop through the key frame labels, and export by interpolating
            # the non key frames.
            start_idx = labels.labels[0].frameindex
            end_idx = len(dp.items_list)
            curr_idx = start_idx  # the current real frame index
            frame_label = labels.labels[curr_idx].category[0][-1]
            key_index = 0  # the current key frame index
            pbar = tqdm(total=end_idx)
            while curr_idx != end_idx:
                # check if frame is key frame, if yes get the new labels, else keep old
                if (
                    key_index < len(labels.labels)
                    and labels.labels[key_index].frameindex == curr_idx
                ):
                    frame_label = labels.labels[key_index].category[0][-1]
                    key_index += 1

                # check export format type
                if self.format == "redbrick-json":
                    output.append(
                        {
                            "url": dp.items_list_not_signed[curr_idx],
                            "class": frame_label,
                            "name": video_name,
                        }
                    )
                else:
                    # export the frame to the relvant folder
                    if not os.path.isdir(
                        os.path.join(self.cache_dir, video_name, frame_label)
                    ):
                        os.makedirs(
                            os.path.join(self.cache_dir, video_name, frame_label)
                        )

                    frame_data = np.flip(url_to_image(dp.items_list[curr_idx]), axis=2)
                    frame_filename = (
                        dp.items_list_not_signed[curr_idx]
                        .replace("/", "_")
                        .replace(":", "_")
                    )
                    cv2.imwrite(
                        os.path.join(
                            self.cache_dir, video_name, frame_label, frame_filename
                        ),
                        frame_data,
                    )

                # update the index
                curr_idx += 1
                pbar.update(1)
            pbar.close()

            if self.format == "redbrick-json":
                with open(os.path.join(self.cache_dir, "output.json"), "w+") as file:
                    json.dump(output, file, indent=2)



"""
Class for controlling image exports.
"""
from redbrick.entity.export.export_base import ExportBase
from dataclasses import dataclass
from termcolor import colored
import os
import cv2  # type: ignore
import json
from tqdm import tqdm  # type: ignore
import numpy as np  # type: ignore
from redbrick.entity.datapoint import Image, Video
import shutil


@dataclass
class ExportImage(ExportBase):
    """Handle image exports."""

    def export(self) -> None:
        """Export the images and labels."""
        print(colored("[INFO]:", "blue"), "Cacheing labels, and data...")
        self.cache()

    def cache(self) -> None:
        """Cache the images and labels."""
        ExportBase.cache(self)

    def cache_bbox(self) -> None:
        """Cache image bounding boxes."""
        # Create YOLO style meta data files
        num_classes = len(list(self.labelset.taxonomy.keys()))
        train_file = "train.txt"
        names_file = "names.txt"
        data_file = "obj.data"
        backup_dir = "backup/"

        # obj.data
        with open(self.cache_dir + "/" + data_file, "w+") as file:
            file.write("classes = %s" % num_classes)
            file.write("\n")
            file.write("train = %s" % ("data" + "/" + train_file))
            file.write("\n")
            file.write("names = %s" % ("data" + "/" + names_file))
            file.write("\n")
            file.write("backup = %s" % backup_dir)

        # names.txt
        with open(self.cache_dir + "/" + names_file, "w+") as file:
            class_names = list(self.labelset.taxonomy.keys())
            for name in class_names:
                file.write(name + "\n")

        # obj_train_data/
        os.mkdir(os.path.join(self.cache_dir, "obj_train_data"))
        taxonomy_mapper = {
            name: idx for idx, name in enumerate(list(self.labelset.taxonomy.keys()))
        }
        image_filepaths = []

        # IMAGE DATA
        for i in tqdm(range(len(self.labelset.dp_ids))):
            dp = self.labelset.__getitem__(i)
            if isinstance(dp, Video):
                # Ensure the data type is Image
                raise ValueError(
                    "Oops, something went wrong! Please reach out to contact@redbrickai.com."
                )
            dp_entry = {}
            dp_entry["url"] = dp.image_url_not_signed
            dp_entry["labels"] = []

            # write image data to file
            image_filepath = os.path.join(
                self.cache_dir,
                "obj_train_data",
                str(dp.image_url_not_signed).replace("/", "_"),
            )
            image_filepaths.append(
                "obj_train_data/" + str(dp.image_url_not_signed).replace("/", "_")
            )
            cv2.imwrite(  # pylint: disable=no-member
                image_filepath, np.flip(dp.image_data, axis=2)
            )

            # create the label file name
            filext_idx = image_filepath.rfind(".")
            if not filext_idx:
                filext_idx = -1
            label_filepath = image_filepath[0:filext_idx] + ".txt"

            # write labels to the txt file
            with open(label_filepath, "w+") as file:
                for label in dp.labels:
                    class_idx = taxonomy_mapper[label.classname[0][-1]]
                    file.write(
                        "%d %.6f %.6f %.6f %.6f \n"
                        % (
                            class_idx,
                            label.xnorm,
                            label.ynorm,
                            label.wnorm,
                            label.hnorm,
                        )
                    )

        # create train.txt file
        with open(os.path.join(self.cache_dir, train_file), "w+") as file:
            for filename in image_filepaths:
                file.write(filename + "\n")

    def cache_segmentation(self) -> None:
        """Cache image segmentation."""
        # Save masks
        for i in tqdm(range(len(self.labelset.dp_ids))):
            dp = self.labelset.__getitem__(i)
            dp.labels.mask.dump(
                self.cache_dir
                + "/"
                + str(dp.image_url_not_signed).replace("/", "_")
                + ".dat"
            )

        # Save the class-mapping (taxonomy) in json format in the folder
        with open("%s/class-mapping.json" % self.cache_dir, "w+") as file:
            json.dump(dp.taxonomy, file, indent=2)

    def cache_classify(self) -> None:
        """Cache image classification."""
        pass



"""
Test the taxonomy2 class.
"""

from redbrick.entity.taxonomy2 import Taxonomy2
import json

# DATA
TAXONOMY1 = {
    "taxonomy": {
        "name": "DEFAULT::YoloThings",
        "version": 1,
        "categories": [
            {
                "name": "object",
                "children": [
                    {"name": "person", "classId": 0, "children": []},
                    {"name": "bicycle", "classId": 1, "children": []},
                    {"name": "car", "classId": 2, "children": []},
                    {"name": "motorbike", "classId": 3, "children": []},
                    {"name": "aeroplane", "classId": 4, "children": []},
                ],
            }
        ],
    }
}
TAXONOMY1_CLASS_ID = {
    "person": 0,
    "bicycle": 1,
    "car": 2,
    "motorbike": 3,
    "aeroplane": 4,
}
TAXONOMY1_ID_CLASS = {
    0: "person",
    1: "bicycle",
    2: "car",
    3: "motorbike",
    4: "aeroplane",
}


def test_always_passes() -> None:
    """Dummmy."""
    assert True


def test_init() -> None:
    """Test correct init."""
    # init
    Tax = Taxonomy2(remote_tax=TAXONOMY1["taxonomy"])

    # dummy check
    assert json.dumps(Tax.taxonomy) == json.dumps(TAXONOMY1["taxonomy"])

    # class/id map check
    assert json.dumps(Tax.taxonomy_class_id_map) == json.dumps(TAXONOMY1_CLASS_ID)
    assert json.dumps(Tax.taxonomy_id_class_map) == json.dumps(TAXONOMY1_ID_CLASS)


# DATA
TAXONOMY2 = {
    "taxonomy": {
        "name": "DEFAULT::YoloThings",
        "version": 1,
        "categories": [
            {
                "name": "object",
                "children": [
                    {
                        "name": "person",
                        "classId": 0,
                        "children": [
                            {
                                "name": "boy",
                                "classId": 5,
                                "children": [
                                    {"name": "young", "classId": 7, "children": []}
                                ],
                            },
                            {"name": "girl", "classId": 6, "children": []},
                        ],
                    },
                    {"name": "bicycle", "classId": 1, "children": []},
                    {"name": "car", "classId": 2, "children": []},
                    {"name": "motorbike", "classId": 3, "children": []},
                    {"name": "aeroplane", "classId": 4, "children": []},
                ],
            }
        ],
    }
}
TAXONOMY2_CLASS_ID = {
    "person": 0,
    "bicycle": 1,
    "car": 2,
    "motorbike": 3,
    "aeroplane": 4,
    "boy": 5,
    "girl": 6,
    "young": 7,
}
TAXONOMY2_ID_CLASS = {
    0: "person",
    1: "bicycle",
    2: "car",
    3: "motorbike",
    4: "aeroplane",
    5: "boy",
    6: "girl",
    7: "young",
}


def test_recursive() -> None:
    """Test tree taxonomy."""
    Tax = Taxonomy2(remote_tax=TAXONOMY2["taxonomy"])

    # Check maps
    assert Tax.taxonomy_id_class_map == TAXONOMY2_ID_CLASS
    assert Tax.taxonomy_class_id_map == TAXONOMY2_CLASS_ID



"""Public interface to remote_label."""
from redbrick.api import RedBrickApi
import json
from redbrick.entity.task import Task
from redbrick.entity.label import ImageBoundingBox, VideoBoundingBox, VideoClassify
from redbrick.entity.taxonomy2 import Taxonomy2
from typing import List, Union
from termcolor import colored


class RemoteLabel:
    """An interface to RemoteLabel brick."""

    def __init__(self, org_id: str, project_id: str, stage_name: str) -> None:
        """Construct RemoteLabel instance."""
        print(colored("[INFO]:", "blue"), "Initializing remote-labeling module...")
        self.org_id = org_id
        self.project_id = project_id
        self.stage_name = stage_name
        self.api_client = RedBrickApi(cache=False)

        # Gather stage information and store
        stage_info = self.api_client.get_stage(
            org_id=org_id, project_id=project_id, stage_name=stage_name
        )
        taxonomy = self.api_client.get_taxonomy(
            orgId=org_id,
            name=stage_info["outputTaxonomyName"],
            version=stage_info["outputTaxonomyVersion"],
        )
        self.taxonomy: Taxonomy2 = taxonomy
        self.task_type = stage_info["outputType"]

    def get_num_tasks(self) -> int:
        """Get the number of tasks queued."""
        num = self.api_client.get_num_remote_tasks(
            org_id=self.org_id, project_id=self.project_id, stage_name=self.stage_name
        )
        return num

    def get_task(self, num_tasks: int) -> List[Task]:
        """User facing function to get task."""
        print(colored("[INFO]:", "blue"), "Retrieving task from backend...", end=" ")
        task = self.__get_remote_labeling_task(num_tasks=num_tasks)
        if len(task) == 0:
            print(colored("\n[WARNING]:", "yellow"), "No more tasks in this stage.")
            return task
        print(colored("Done.", "green"))
        return task

    def submit_task(
        self,
        task: Task,
        labels: Union[ImageBoundingBox, VideoBoundingBox, VideoClassify],
    ) -> None:
        """User facing funciton to submit a task."""
        print(colored("[INFO]", "blue"), "Submitting task to backend...", end=" ")
        new_subname = "remote-labeling"

        # Check that label category matches taxonomy
        check, classname = labels.compare_taxonomy(taxonomy=self.taxonomy)
        if not check:
            raise ValueError(
                "%s is not a valid category for taxonomy %s"
                % (classname, self.taxonomy.name)
            )

        # Check label type
        if task.task_data_type == "IMAGE_BBOX":
            if not isinstance(labels, ImageBoundingBox):
                raise ValueError("Labels must be of type ImageBoundingBox!")
        if task.task_data_type == "VIDEO_BBOX":
            if not isinstance(labels, VideoBoundingBox):
                raise ValueError("Labels must be of type VideoBoundingBox!")

        # Put task data
        self.__put_task_data(
            dp_id=task.dp_id,
            sub_name=new_subname,
            task_data=labels,
            taxonomy_name=task.taxonomy["name"],
            taxonomy_version=task.taxonomy["version"],
            td_type=task.task_data_type,
        )

        # Put remote labeling task
        submit_task = Task(
            org_id=task.org_id,
            project_id=task.project_id,
            stage_name=task.stage_name,
            task_id=task.task_id,
            dp_id=task.dp_id,
            sub_name=new_subname,
            taxonomy=task.taxonomy,
            items_list=task.items_list,
            items_list_presigned=task.items_list_presigned,
            task_data_type=task.task_data_type,
        )
        self.__put_remote_labeling_task(submit_task)

        print(colored("Done.", "green"))

    def __put_task_data(
        self,
        dp_id: str,
        sub_name: str,
        task_data: Union[ImageBoundingBox, VideoBoundingBox, VideoClassify],
        taxonomy_name: str,
        taxonomy_version: str,
        td_type: str,
    ) -> None:
        """Read labels from local folder, and submit the labels."""
        task_datas = str(task_data)  # Stringify the json object

        self.api_client.putTaskData(
            org_id=self.org_id,
            project_id=self.project_id,
            dp_id=dp_id,
            stage_name=self.stage_name,
            sub_name=sub_name,
            task_data=task_datas,
            taxonomy_name=taxonomy_name,
            taxonomy_version=taxonomy_version,
            td_type=td_type,
        )

    def __put_remote_labeling_task(self, task: Task) -> None:
        """Put the remote labeling task to the backend."""
        finishedTask = {
            "orgId": task.org_id,
            "projectId": task.project_id,
            "stageName": task.stage_name,
            "taskId": task.task_id,
            "newSubName": task.sub_name,
        }
        self.api_client.putRemoteLabelingTask(finishedTask)

    def __get_remote_labeling_task(self, num_tasks: int) -> List[Task]:
        """Get the labeling tasks from API."""
        task = self.api_client.tasksToLabelRemote(
            orgId=self.org_id,
            projectId=self.project_id,
            stageName=self.stage_name,
            numTasks=num_tasks,
        )
        return task



"""RedBrick Client."""

from typing import Optional, Any


class RedBrickClient:
    """Interface to RedBrick platform."""

    class __RedBrickClient:
        """A private singleton for redbrick."""

        def __init__(self, api_key: str, custom_url: Optional[str] = None) -> None:
            """Construct RedBrick client singleton."""
            self.api_key = api_key
            self.custom_url = custom_url

        def __str__(self) -> str:
            """Get string representation."""
            return repr(self) + "***" + self.api_key[-4:-1]

    instance: Optional[__RedBrickClient] = None

    def __init__(
        self, api_key: Optional[str] = None, url: Optional[str] = None
    ) -> None:
        """Construct instance of RedBrickClient."""
        if not api_key:
            if not RedBrickClient.instance:
                raise Exception("Must specify api_key")
            return
        if not RedBrickClient.instance:
            RedBrickClient.instance = RedBrickClient.__RedBrickClient(api_key, url)
        else:
            RedBrickClient.instance.api_key = api_key

    def __getattr__(self, name: str) -> Any:
        """Get specified attribute from object."""
        return getattr(self.instance, name)



"""A higher level abstraction."""

from typing import Optional, List, Union, Dict
from random import randint

import numpy as np  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
from matplotlib import patches
import os
import datetime
from tqdm import tqdm  # type: ignore
import logging
from termcolor import colored
import json
import cv2  # type: ignore
import random

from redbrick.labelset.labelset_base import LabelsetBase
from redbrick.api import RedBrickApi
from redbrick.entity.datapoint import Image, Video
from redbrick.entity.export import ExportImage, ExportVideo


class LabelsetLoader(LabelsetBase):
    """A basic high level loader class."""

    def __init__(self, org_id: str, label_set_name: str) -> None:
        """Construct Loader."""
        self.org_id = org_id
        self.label_set_name = label_set_name
        self.api_client = RedBrickApi(cache=False)

        print(colored("[INFO]:", "blue"), "Counting available datapoints...")
        # All datapoints in labelset
        self.dp_ids, custom_group = self.api_client.get_datapoint_ids(
            self.org_id, self.label_set_name
        )
        self.task_type = custom_group.task_type
        self.data_type = custom_group.data_type
        self.taxonomy: Dict[str, int] = custom_group.taxonomy
        print(
            colored("[INFO]:", "blue"), "Number of Datapoints = %s" % len(self.dp_ids)
        )

        # Update taxonomy mapper if segmentation
        if self.task_type == "SEGMENTATION":
            self.taxonomy_update_segmentation()

    def __getitem__(self, index: int) -> Union[Image, Video]:
        """Get information needed for a single item."""
        dp = self.api_client.get_datapoint(
            self.org_id,
            self.label_set_name,
            self.dp_ids[index],
            self.task_type,
            self.taxonomy,
        )
        return dp

    def export(self, format: str = "redbrick") -> str:
        """Export."""

        if self.data_type == "IMAGE":
            export_img: ExportImage = ExportImage(format=format, labelset=self)
            export_img.export()
            return export_img.cache_dir
        elif self.data_type == "VIDEO":
            export_vid: ExportVideo = ExportVideo(format=format, labelset=self)
            export_vid.export()
            return export_vid.cache_dir
        else:
            raise ValueError(
                "%s data type not supported! Please reach out to contact@redbrickai.com"
                % self.data_type
            )

    def number_of_datapoints(self) -> int:
        """Get number of datapoints."""
        return len(self.dp_ids)

    def show_data(self) -> None:
        """Visualize the data."""

        if self.data_type == "VIDEO":
            print(colored("[INFO]:", "blue"), "Visualizing first 20 frames...")

            num_dps = self.number_of_datapoints()

            self[0].show_data()
            return

        # Image data type
        print(colored("[INFO]:", "blue"), "Visualizing data and labels...")

        # Prepare figure
        num_dps = self.number_of_datapoints()
        cols = int(np.min([2, num_dps]))
        rows = int(np.min([2, np.ceil(num_dps / cols)]))
        fig = plt.figure()

        # Generate random index list
        list_len = np.min([rows * cols, num_dps])
        indexes = random.sample(range(0, list_len), list_len)

        # Iterate through axes
        for i, idx in enumerate(indexes):

            ax = fig.add_subplot(rows, cols, i + 1)
            self[idx].show_data(ax=ax)

        plt.tight_layout()
        plt.show()

    def taxonomy_update_segmentation(self) -> None:
        """Fix the taxonomy mapper object to be 1-indexed for segmentation projects."""
        for key in self.taxonomy.keys():
            self.taxonomy[key] += 1
            if self.taxonomy[key] == 0:
                print(
                    colored("[ERROR]:", "red"),
                    "Taxonomy class id's must be 0 indexed. Please contact contact@redbrickai.com for help.",
                )
                exit(1)

        # Add a background class for segmentation
        self.taxonomy["background"] = 0



"""
Representation for the video datapoint type.
"""

from dataclasses import dataclass, field
from .base_datapoint import BaseDatapoint
from typing import List, Union, Any, Optional
from redbrick.entity.label import VideoClassify, VideoBoundingBox
from redbrick.utils import url_to_image
import matplotlib.pyplot as plt  # type: ignore
import matplotlib.animation as animation  # type: ignore
from tqdm import tqdm  # type: ignore


@dataclass
class Video(BaseDatapoint):
    """Video object."""

    dp_id: str
    video_name: str
    items_list: List[str]
    items_list_not_signed: List[str]
    labels: Union[VideoBoundingBox, VideoClassify] = field(init=False)

    def __post_init__(self) -> None:
        """Run after init."""
        if self.task_type == "BBOX":
            self.labels = VideoBoundingBox(labels=self.remote_labels)
        elif self.task_type == "CLASSIFY":
            self.labels = VideoClassify(remote_labels=self.remote_labels)
        else:
            raise ValueError(
                "%s task type not supported! Please reach out to contact@redbrickai.com."
                % self.task_type
            )

    def show_data(self, start: int = 0, end: int = 20) -> None:
        """Show the video data."""
        ims = []
        fig, ax = plt.subplots()
        num_frames = len(self.items_list)

        for i in tqdm(range(start, min(end, num_frames))):
            frame = url_to_image(url=self.items_list[i])
            height, width, _ = frame.shape
            im = plt.imshow(frame, animated=True)

            ims_ = self.labels.show(
                ax=ax,
                width=width,
                height=height,
                frameindex=i,
                num_frames=len(self.items_list),
            )

            ims_ = [im] + ims_  # the frame data im needs to be first
            ims.append(ims_)

        animation.ArtistAnimation(
            fig, ims, interval=50, blit=True, repeat_delay=1000,
        )
        plt.show()



"""
Representation of classification labels.
"""

from dataclasses import dataclass, field
from redbrick.entity.taxonomy2 import Taxonomy2
from redbrick.utils import compare_taxonomy
from typing import List, Union, Tuple, Dict, Any
import json
import uuid


@dataclass
class VideoClassifyEntry:
    category: List[List[str]]
    labelid: str
    frameclassify: bool
    trackid: str
    keyframe: bool
    end: bool
    frameindex: int


@dataclass
class ImageClassifyEntry:
    category: List[List[str]]


@dataclass
class ImageClassify:
    remote_labels: List[Any]
    labels: ImageClassifyEntry = field(init=False)

    def __post_init__(self) -> None:
        """Run after init."""
        # TODO
        pass

    def compare_taxonomy(self, taxonomy: Taxonomy2) -> Tuple[bool, List[List[str]]]:
        """Ensure self labels match taxonomy."""
        check = compare_taxonomy(category_path=self.labels.category, taxonomy=taxonomy)

        if check:
            return check, [["null", "null"]]
        else:
            return check, self.labels.category

    def show(self, ax: Any, width: int, height: int) -> None:
        """Show the image classify labels."""
        raise NotImplementedError()


@dataclass
class VideoClassify:
    remote_labels: List[Any]
    labels: List[VideoClassifyEntry] = field(init=False)

    def __post_init__(self) -> None:
        """After init."""
        entries = []
        trackid = str(uuid.uuid4())
        for label in self.remote_labels:

            # Fill out values if not provided
            if not "trackid" in label:
                label["trackid"] = trackid
            if not "end" in label:
                label["end"] = True
            if not "keyframe" in label:
                label["keyframe"] = True
            if not "labelid" in label:
                label["labelid"] = str(uuid.uuid4())

            entry = VideoClassifyEntry(
                category=label["category"],
                labelid=label["labelid"],
                frameclassify=label["frameclassify"],
                trackid=label["trackid"],
                keyframe=label["keyframe"],
                end=label["end"],
                frameindex=label["frameindex"],
            )
            entries.append(entry)

        self.labels = entries

    def compare_taxonomy(self, taxonomy: Taxonomy2) -> Tuple[bool, List[List[str]]]:
        """Ensure self labels match taxonomy."""

        for label in self.labels:
            check = compare_taxonomy(category_path=label.category, taxonomy=taxonomy)
            if not check:
                return check, label.category

        return True, [["null", "null"]]

    def __str__(self) -> str:
        """String representation of video classification."""
        labels = []
        for label in self.labels:
            entry: Dict[Any, Any] = {}
            entry["category"] = label.category
            entry["keyframe"] = label.keyframe
            entry["frameclassify"] = label.frameclassify
            entry["end"] = label.end
            entry["labelid"] = label.labelid
            entry["frameindex"] = label.frameindex
            entry["trackid"] = label.trackid
            entry["attributes"] = []

            labels.append(entry)

        output = {"labels": labels}
        return json.dumps(output)

    def show(
        self, ax: Any, width: int, height: int, frameindex: int, num_frames: int
    ) -> List[Any]:
        """show the labels."""
        raise NotImplementedError()



# mypy: ignore-errors
# pylint: skip-file

"""
Test the labelset module.
"""

import redbrick
import unittest
from typing import Any
from unittest.mock import patch, MagicMock
from redbrick.entity.custom_group import CustomGroup
from redbrick.tests.taxonomy2_test import TAXONOMY1
from .datapoint_test import IMAGE, VIDEO, VIDEO_CLASSIFY
import os
import numpy as np
import shutil
import json
import uuid


# d888888b .88b  d88.  .d8b.   d888b  d88888b
#   `88'   88'YbdP`88 d8' `8b 88' Y8b 88'
#    88    88  88  88 88ooo88 88      88ooooo
#    88    88  88  88 88~~~88 88  ooo 88~~~~~
#   .88.   88  88  88 88   88 88. ~8~ 88.
# Y888888P YP  YP  YP YP   YP  Y888P  Y88888P


class Mock_Image:
    """Mocking RedBrickApi for images."""

    def __init__(self):
        return

    def get_datapoint_ids(self, org_id, label_set_name):
        return (
            ["1", "2"],
            CustomGroup("BBOX", "IMAGE", TAXONOMY1["taxonomy"]),
        )

    def get_datapoint(self, orgid, labelsetname, dpid, tasktype, taxonomy):
        return IMAGE


@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_init(mock: Any) -> None:
    """Test init of labelset."""

    mock.return_value = Mock_Image()
    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    assert labelset.task_type == "BBOX"
    assert len(labelset.dp_ids) == 2


@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_export_image(mock: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Image()

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    cache_dir = labelset.export(format="redbrick")

    assert os.path.isdir(cache_dir)
    assert os.listdir(os.path.join(cache_dir, "obj_train_data")) == [
        IMAGE.image_url_not_signed[0:-4] + ".png",
        IMAGE.image_url_not_signed[0:-4] + ".txt",
    ]
    shutil.rmtree(cache_dir)


# db    db d888888b d8888b. d88888b  .d88b.
# 88    88   `88'   88  `8D 88'     .8P  Y8.
# Y8    8P    88    88   88 88ooooo 88    88
# `8b  d8'    88    88   88 88~~~~~ 88    88
#  `8bd8'    .88.   88  .8D 88.     `8b  d8'
#    YP    Y888888P Y8888D' Y88888P  `Y88P'


class Mock_Video:
    """Mocking RedBrickApi for video."""

    def __init__(self):
        return

    def get_datapoint_ids(self, org_id, label_set_name):
        return (
            ["1", "2"],
            CustomGroup("BBOX", "VIDEO", TAXONOMY1["taxonomy"]),
        )

    def get_datapoint(self, orgid, labelsetname, dpid, tasktype, taxonomy):
        return VIDEO


class Mock_Video_Classify:
    """Mocking RedBrickApi for video."""

    def __init__(self):
        return

    def get_datapoint_ids(self, org_id, label_set_name):
        return (
            ["1", "2"],
            CustomGroup("CLASSIFY", "VIDEO", TAXONOMY1["taxonomy"]),
        )

    def get_datapoint(self, orgid, labelsetname, dpid, tasktype, taxonomy):
        VIDEO_CLASSIFY.video_name = dpid
        return VIDEO_CLASSIFY


@patch("redbrick.entity.export.export_video.url_to_image")
@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_export_image(mock: Any, url_to_image: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Video()
    url_to_image.return_value = np.zeros((10, 10, 3))

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    cache_dir = labelset.export(format="redbrick")

    assert os.path.isdir(cache_dir)

    outdir = []
    for item in VIDEO.items_list_not_signed:
        outdir += [(item[0:-4] + ".txt").replace("/", "_").replace(":", "_")]
        outdir += [(item[0:-4] + ".png").replace("/", "_").replace(":", "_")]

    assert os.listdir(os.path.join(cache_dir, "obj_train_data")).sort() == outdir.sort()
    shutil.rmtree(cache_dir)


@patch("redbrick.entity.export.export_video.url_to_image")
@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_export_video_clasify(mock: Any, url_to_image: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Video_Classify()
    url_to_image.return_value = np.zeros((10, 10, 3))

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    cache_dir = labelset.export(format="redbrick")

    assert os.path.isdir(cache_dir)

    files = os.listdir(cache_dir)
    assert len(files) == 2
    assert files == ["1", "2"]
    shutil.rmtree(cache_dir)


@patch("redbrick.entity.export.export_video.url_to_image")
@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_export_video_clasify_json(mock: Any, url_to_image: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Video_Classify()
    url_to_image.return_value = np.zeros((10, 10, 3))

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    cache_dir = labelset.export(format="redbrick-json")

    assert os.path.isdir(cache_dir)

    files = os.listdir(cache_dir)
    assert len(files) == 1

    # open the json file
    with open(os.path.join(cache_dir, files[0]), "r") as file:
        labels = json.load(file)

    shutil.rmtree(cache_dir)



"""
Initialize RedBrick package.

Derek Lukacs, 2020
"""

from .client import RedBrickClient
import redbrick.api
from typing import Optional

# import redbrick.base
import redbrick.client
import redbrick.labelset
import redbrick.remote_label


def init(api_key: str, url: Optional[str] = None) -> None:
    """Initialize package state."""
    RedBrickClient(api_key, url)



"""
Class representation of a task.
"""
from dataclasses import dataclass
from typing import Dict, Any
from redbrick.utils import url_to_image
import numpy as np  # type: ignore


@dataclass
class Task:
    """Class representation of a task."""

    org_id: str
    project_id: str
    stage_name: str
    task_id: str
    dp_id: str
    sub_name: str
    taxonomy: Dict[Any, Any]
    items_list: str
    items_list_presigned: str
    task_data_type: str

    @staticmethod
    def get_data(presigned_url: str) -> np.ndarray:
        """Get the frame/image data."""
        return url_to_image(presigned_url)



"""
Test the bounding box module.
"""
import pytest
from redbrick.entity.label import ImageBoundingBox, VideoBoundingBox
from redbrick.entity.taxonomy2 import Taxonomy2
from redbrick.tests.taxonomy2_test import TAXONOMY2, TAXONOMY1
import json
import uuid

# d888888b .88b  d88.  .d8b.   d888b  d88888b   d8888b.  .d8b.  d888888b  .d8b.
#   `88'   88'YbdP`88 d8' `8b 88' Y8b 88'       88  `8D d8' `8b `~~88~~' d8' `8b
#    88    88  88  88 88ooo88 88      88ooooo   88   88 88ooo88    88    88ooo88
#    88    88  88  88 88~~~88 88  ooo 88~~~~~   88   88 88~~~88    88    88~~~88
#   .88.   88  88  88 88   88 88. ~8~ 88.       88  .8D 88   88    88    88   88
# Y888888P YP  YP  YP YP   YP  Y888P  Y88888P   Y8888D' YP   YP    YP    YP   YP

LABEL1 = [
    {
        "category": [["object", "car"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
    }
]

LABEL2 = [
    {
        "category": [["object", "person"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
    },
    {
        "category": [["object", "car"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
    },
]

LABEL3 = [
    {
        "category": [["object", "person"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
    },
    {
        "category": [["object", "not correct"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
    },
]


def test_init_image_bbox() -> None:
    """Test initializing."""
    image_bbox = ImageBoundingBox(labels=LABEL1)
    assert len(image_bbox.labels) == 1
    assert image_bbox.labels[0].xnorm == 0.1
    assert image_bbox.labels[0].ynorm == 0.1
    assert image_bbox.labels[0].hnorm == 0.1
    assert image_bbox.labels[0].wnorm == 0.1
    assert image_bbox.labels[0].classname == [["object", "car"]]

    image_bbox2 = ImageBoundingBox(labels=LABEL2)
    assert len(image_bbox2.labels) == 2
    assert image_bbox2.labels[0].classname == [["object", "person"]]


def test_compare_tax_image_bbox() -> None:
    """Test comparison of taxonomy."""
    taxonomy1 = Taxonomy2(remote_tax=TAXONOMY1["taxonomy"])

    image_bbox = ImageBoundingBox(labels=LABEL1)
    assert image_bbox.compare_taxonomy(taxonomy=taxonomy1)[0]

    image_bbox2 = ImageBoundingBox(labels=LABEL3)
    assert not image_bbox2.compare_taxonomy(taxonomy=taxonomy1)[0]


# db    db d888888b d8888b. d88888b  .d88b.    d8888b.  .d8b.  d888888b  .d8b.
# 88    88   `88'   88  `8D 88'     .8P  Y8.   88  `8D d8' `8b `~~88~~' d8' `8b
# Y8    8P    88    88   88 88ooooo 88    88   88   88 88ooo88    88    88ooo88
# `8b  d8'    88    88   88 88~~~~~ 88    88   88   88 88~~~88    88    88~~~88
#  `8bd8'    .88.   88  .8D 88.     `8b  d8'   88  .8D 88   88    88    88   88
#    YP    Y888888P Y8888D' Y88888P  `Y88P'    Y8888D' YP   YP    YP    YP   YP

VID_LABEL1 = [
    {
        "category": [["object", "person"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
        "frameindex": 0,
    },
    {
        "category": [["object", "car"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
        "frameindex": 0,
    },
    {
        "category": [["object", "car"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
        "frameindex": 1,
    },
]


def test_vid_bbox() -> None:
    """Test vid bbox init."""

    vid_bbox = VideoBoundingBox(labels=VID_LABEL1)

    assert len(vid_bbox.labels) == 3
    assert len(vid_bbox.labels[0].labelid) == 36
    assert vid_bbox.labels[1].keyframe


def test_vid_bbox_str() -> None:
    """Test string conversion."""

    vid_bbox = VideoBoundingBox(labels=VID_LABEL1)

    output = {
        "labels": [
            {
                "category": [["object", "person"]],
                "attributes": [],
                "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
                "frameindex": 0,
                "labelid": vid_bbox.labels[0].labelid,
                "keyframe": True,
                "end": True,
                "trackid": vid_bbox.labels[0].trackid,
            },
            {
                "category": [["object", "car"]],
                "attributes": [],
                "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
                "frameindex": 0,
                "labelid": vid_bbox.labels[1].labelid,
                "keyframe": True,
                "end": True,
                "trackid": vid_bbox.labels[1].trackid,
            },
            {
                "category": [["object", "car"]],
                "attributes": [],
                "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
                "frameindex": 1,
                "labelid": vid_bbox.labels[2].labelid,
                "keyframe": True,
                "end": True,
                "trackid": vid_bbox.labels[2].trackid,
            },
        ],
    }

    assert json.loads(str(vid_bbox)) == output


t1 = str(uuid.uuid4())
t2 = str(uuid.uuid4())


def test_interpolate_one_frame() -> None:
    """Test interpolating only one frame."""
    # arrange
    label = {
        "category": [["object", "person"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
        "frameindex": 0,
        "trackid": t1,
        "keyframe": True,
        "end": True,
    }
    vid_bbox = VideoBoundingBox(labels=[label])
    # assume
    assert len(vid_bbox.labels) == 1

    # action
    result = vid_bbox.interpolate_labels(num_frames=1)

    # assert
    assert len(result) == 1
    assert len(result[0]) == 1
    result_label = result[0][0]
    assert result_label.xnorm == 0.1
    assert result_label.keyframe
    assert result_label.end


def test_interpolate_one_end_label_10_frames() -> None:
    # arrange
    label = {
        "category": [["object", "person"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
        "frameindex": 0,
        "trackid": t1,
        "keyframe": True,
        "end": True,
    }
    vid_bbox = VideoBoundingBox(labels=[label])
    # assume
    assert len(vid_bbox.labels) == 1

    # action
    result = vid_bbox.interpolate_labels(num_frames=10)

    # assert
    assert len(result) == 10
    assert len(result[0]) == 1
    assert len(result[1]) == 0
    result_label = result[0][0]
    assert result_label.xnorm == 0.1
    assert result_label.keyframe
    assert result_label.end


def test_interpolate_one_label_10_frames() -> None:
    # arrange
    label = {
        "category": [["object", "person"]],
        "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
        "frameindex": 0,
        "trackid": t1,
        "keyframe": True,
        "end": False,
    }
    vid_bbox = VideoBoundingBox(labels=[label])
    # assume
    assert len(vid_bbox.labels) == 1, vid_bbox.labels

    # action
    result = vid_bbox.interpolate_labels(num_frames=10)

    # assert
    assert len(result) == 10
    assert len(result[0]) == 1
    result_label = result[0][0]

    assert result_label.xnorm == 0.1
    assert result_label.keyframe
    assert result_label.end == False
    last_label = result[-1][0]
    assert last_label.xnorm == 0.1
    assert last_label.end == False
    assert last_label.keyframe == False


def test_interpolate_two_labels() -> None:
    # arrange
    VID_LABEL2 = [
        {
            "category": [["object", "person"]],
            "bbox2d": {"xnorm": 0.1, "ynorm": 0.1, "hnorm": 0.1, "wnorm": 0.1,},
            "frameindex": 1,
            "trackid": t1,
            "keyframe": True,
            "end": False,
        },
        {
            "category": [["object", "person"]],
            "bbox2d": {"xnorm": 0.5, "ynorm": 0.5, "hnorm": 0.1, "wnorm": 0.5,},
            "frameindex": 3,
            "trackid": t1,
            "keyframe": True,
            "end": True,
        },
    ]
    vid_bbox = VideoBoundingBox(labels=VID_LABEL2)
    # assume
    assert len(vid_bbox.labels) == 2, vid_bbox.labels

    # acction
    result = vid_bbox.interpolate_labels(num_frames=10)

    # assert
    assert len(result) == 10
    assert len(result[0]) == 0

    assert len(result[1]) == 1
    assert result[1][0].keyframe

    assert len(result[2]) == 1
    assert not result[2][0].keyframe
    assert result[2][0].xnorm == 0.3

    assert len(result[3]) == 1
    assert result[3][0].keyframe

    assert len(result[4]) == 0



"""
Representation for the image datapoint type.
"""
from dataclasses import dataclass, field
from .base_datapoint import BaseDatapoint
import numpy as np  # type: ignore
from typing import Union, Any
import matplotlib.pyplot as plt  # type: ignore
from redbrick.entity.label import ImageBoundingBox, ImageClassify, ImageSegmentation


@dataclass
class Image(BaseDatapoint):
    """Image object."""

    dp_id: str
    image_url: str
    image_url_not_signed: str
    image_data: np.ndarray
    labels: Union[ImageBoundingBox, ImageClassify, ImageSegmentation] = field(
        init=False
    )

    def __post_init__(self) -> None:
        """Run after init."""
        if self.task_type == "BBOX":
            self.labels = ImageBoundingBox(labels=self.remote_labels)
        elif self.task_type == "CLASSIFY":
            self.labels = ImageClassify(remote_labels=self.remote_labels)
        elif self.task_type == "SEGMENTATION":
            self.labels = ImageSegmentation(
                remote_labels=self.remote_labels, classes=self.taxonomy
            )
        else:
            raise ValueError(
                "%s task type is not supported. Please reach out to contact@redbrickai.com"
                % self.task_type
            )

    def show_data(self, ax: Any = None) -> None:
        """Show the data with the ground truth."""
        if ax is None:
            _, ax = plt.subplots()

        # Show the image
        ax.imshow(self.image_data)

        # Render labels
        height, width, _ = self.image_data.shape
        self.labels.show(ax=ax, width=width, height=height)
        title = (
            "..." + self.image_url_not_signed[-10:]
            if len(self.image_url_not_signed) > 10
            else self.image_url_not_signed
        )
        ax.set_title("%s" % title)

        if ax is None:
            plt.show()



from setuptools import setup, find_packages  # type: ignore

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="redbrick-sdk",
    url="https://github.com/dereklukacs/redbrick-sdk",
    version="0.2.8",
    description="RedBrick platform python SDK!",
    py_modules=["redbrick"],
    packages=find_packages(),
    # package_dir={"": "redbrick"},
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "opencv-python==4.2.0.34",
        "numpy==1.18.5",
        "matplotlib==3.2.1",
        "requests==2.23.0",
        "tqdm==4.50.0",
        "scikit-image==0.17.1",
        "termcolor==1.1.0",
    ],
    extras_require={
        "dev": [
            "pytest>=3.7",
            "black==19.10b0",
            "pydocstyle==5.0.2",
            "pycodestyle==2.6.0",
            "twine==3.1.1",
            "wheel==0.35.1",
            "mypy==0.790",
            "pytest-mock==3.3.1",
        ]
    },
)



"""
Representation of a segmentation label.
"""

from dataclasses import dataclass, field
from typing import Dict, List, Any, Optional
import numpy as np  # type: ignore
import json
import copy
from skimage import draw  # type: ignore
import matplotlib.cm  # type: ignore


@dataclass
class ImageSegmentation:
    classes: Dict[str, int]
    remote_labels: List[Any]

    url: str = field(init=False)
    mask: np.ndarray = field(init=False)

    def __post_init__(self) -> None:
        """After init."""
        # TODO: Check this code for scenario with no label
        imagesize = self.remote_labels[0]["pixel"]["imagesize"]
        mask = np.zeros([imagesize[1], imagesize[0]])

        with open("debug.json", "w+") as file:
            json.dump(self.remote_labels, file, indent=2)

        for label_obj in self.remote_labels:
            label = label_obj["pixel"]
            category_name = label_obj["category"][0][-1]
            class_id = self.classes[category_name]
            regions = copy.deepcopy(label["regions"])
            holes = copy.deepcopy(label["holes"])
            imagesize = label["imagesize"]

            # Iterate through the regions and create mask
            if regions and len(regions):
                for region in regions:
                    if len(np.array(region).shape) == 1:
                        # Don't add empty regions to the mask
                        break

                    # [x,y] needs to correspond to [r,c]
                    region = np.flip(region, axis=1)

                    mask_ = (
                        draw.polygon2mask([imagesize[1], imagesize[0]], region).astype(
                            float
                        )
                        * class_id
                    )

                    # Add on new regions to root mask object
                    class_id_indexes = np.where(mask_ == class_id)
                    mask[class_id_indexes] = class_id

            # Iterate through the holes and create the negative mask
            neg_mask = np.zeros([imagesize[1], imagesize[0]])
            if holes and len(holes):
                for hole in holes:
                    if len(np.array(hole).shape) == 1:
                        # Don't add empty hole to the negative mask
                        break

                    # [x,y] needs to correspond to [r,c]
                    hole = np.flip(hole, axis=1)

                    neg_mask_ = (
                        draw.polygon2mask([imagesize[1], imagesize[0]], hole).astype(
                            float
                        )
                        * class_id
                    )

                    # Add on new holes to negative mask
                    neg_mask += neg_mask_

            mask -= neg_mask

        # Subtrack out the holes from the region mask
        self.mask = mask

    def color_mask(self, color_map: Any) -> np.ndarray:
        """Return a RGB colored mask."""
        mask_class_ids = np.unique(self.mask)
        color_mask = np.zeros([self.mask.shape[0], self.mask.shape[1], 3])

        # Loop through class range
        for id_ in mask_class_ids:
            if id_ == 0:
                # Do not color class id 0 i.e. the background
                continue

            pixel_class_index = np.where(self.mask == id_)

            # generate colors
            color_mask[pixel_class_index] = np.array(color_map(int(id_))[0:3]) * 256

        color_mask /= 256
        return color_mask

    def show(
        self, ax: Any = None, width: Optional[int] = None, height: Optional[int] = None
    ) -> None:
        """Show the segment labels."""
        cmap = matplotlib.cm.get_cmap("tab10")
        mask = self.color_mask(color_map=cmap)
        ax.imshow(mask, alpha=0.3)



"""
Test the datapoint module.
"""

from redbrick.entity.datapoint import Image, Video
from redbrick.entity.label import ImageBoundingBox, VideoBoundingBox, VideoClassify
from .taxonomy2_test import TAXONOMY1
from .bbox_test import LABEL1, VID_LABEL1
from .classify_test import VID_CLASS_1
import numpy as np  # type: ignore

IMAGE = Image(
    org_id="123",
    label_set_name="abc",
    taxonomy={"person": 0, "car": 1},
    task_type="BBOX",
    remote_labels=LABEL1,
    dp_id="123",
    image_url="abc.png",
    image_url_not_signed="abc234.png",
    image_data=np.zeros((10, 10, 3)),
)

VIDEO = Video(
    org_id="123",
    label_set_name="abc",
    taxonomy={"person": 0, "car": 1},
    task_type="BBOX",
    remote_labels=VID_LABEL1,
    items_list=[
        "http://127.0.0.1:8080/firstadsfsadf.png",
        "http://127.0.0.1:8080/secondasdfdasrqew.png",
        "http://127.0.0.1:8080/thirdadsfqer.png",
    ],
    dp_id="sdfalsn",
    video_name="video name",
    items_list_not_signed=[
        "http://127.0.0.1:8080/firstadsfsadf.png",
        "http://127.0.0.1:8080/secondasdfdasrqew.png",
        "http://127.0.0.1:8080/thirdadsfqer.png",
    ],
)

VIDEO_CLASSIFY = Video(
    org_id="123",
    label_set_name="abc",
    taxonomy={"person": 0, "car": 1},
    task_type="CLASSIFY",
    remote_labels=VID_CLASS_1,
    items_list=[
        "http://127.0.0.1:8080/firstadsfsadf.png",
        "http://127.0.0.1:8080/secondasdfdasrqew.png",
        "http://127.0.0.1:8080/thirdadsfqer.png",
    ],
    dp_id="sdfalsn",
    video_name="video name",
    items_list_not_signed=[
        "http://127.0.0.1:8080/firstadsfsadf.png",
        "http://127.0.0.1:8080/secondasdfdasrqew.png",
        "http://127.0.0.1:8080/thirdadsfqer.png",
    ],
)


def test_image_init() -> None:
    """test init of image."""

    image = IMAGE
    assert len(image.labels.labels) == 1
    assert isinstance(image.labels, ImageBoundingBox)


def test_video_bbox_init() -> None:
    """test init of video."""
    video = VIDEO

    assert len(video.labels.labels) == 3
    assert isinstance(video.labels, VideoBoundingBox)


def test_video_classify_init() -> None:
    """Test video classify."""
    video = VIDEO_CLASSIFY

    assert len(video.labels.labels) == 3
    assert isinstance(video.labels, VideoClassify)
